#!/usr/bin/python3 -u
# -*- coding: utf-8 -*-

from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib
import dbus
import dbus.service
import os
import argparse
import sys
import time

# Victron packages
sys.path.insert(1, os.path.join(os.path.dirname(__file__), './ext/velib_python'))
from dbusmonitor import DbusMonitor
from logger import setup_logging

# configuration
SOFTWARE_VERSION = '0.1'
# devices to ignore on dbus
EXCLUDE_DEVICES = [ 'com.victronenergy.battery.aggregate' ]
# maximum allowed inverter power in Watts
MAX_INVERTER_POWER = 11000
# number of inverter phases present
PHASE_COUNT = 3
# target grid power (if battery state allows it)
GRID_SETPOINT = -20
# threshold when PV power is considered usable
PV_THRESHOLD = 300
# minimum allowed battery SoC
BATTERY_MIN_SOC = 10
# MIN_SOC has to be exceeded by this threshold to enable discharging again
BATTERY_THRESHOLD = 10
# minimum watts to charge from PV when battery is low
BATTERY_LOW_CHARGE = 250
# seconds between inverter/charger mode updates
MODE_UPDATE_DELAY = 5
# feed in excess PV power
FEED_IN_EXCESS = False
# total watts when the fans should switch on
FAN_ON_POWER = 4000
# time to wait before switching off cooling fans
FAN_OFF_DELAY = 900
# how many samples to take for system efficiency average
EFFICIENCY_SAMPLES = 10

class DBusCustomESS:
	def __init__(self):
		# globals
		self._outputEnabled = True
		self._outputLimits = [ -MAX_INVERTER_POWER, 0 ]
		self._batteryLow = False
		self._lastFanStart = 0
		self._lastModeUpdate = 0
		self._efficiency = [0.9] * EFFICIENCY_SAMPLES
		self._efficiencyPos = 0

		# open dbus connection
		dbus.SessionBus() if 'DBUS_SESSION_BUS_ADDRESS' in os.environ else dbus.SystemBus()

		# define which devices and paths to monitor on dbus
		dummy = {'code': None, 'whenToLog': 'configChange', 'accessLevel': None}
		self._dbus_tree = {
			'com.victronenergy.battery': {
				'/Dc/0/Current': dummy, # battery current (+in/-out)
				'/Dc/0/Voltage': dummy, # battery voltage (for power calculations)
				'/Info/MaxChargeCurrent': dummy, # max allowed battery charge current
				'/Info/MaxDischargeCurrent': dummy, # max allowed battery discharge current				
			},
			'com.victronenergy.grid': {
				'/Ac/Power': dummy, # grid meter total power
				'/Role': dummy, # grid meter role, should be 'grid'
			},
			'com.victronenergy.settings': {
				'/Settings/CGwacs/BatteryLife/State': dummy, # BatteryLife, we want 10 for OFF
				'/Settings/CGwacs/Hub4Mode': dummy, # ESS mode, we want mode 3 for manual control
			},
			'com.victronenergy.system': {
				'/Ac/Consumption/L1/Power': dummy, # consumption L1
				'/Ac/Consumption/L2/Power': dummy, # consumption L2
				'/Ac/Consumption/L3/Power': dummy, # consumption L3
				'/Dc/Battery/Power': dummy, # combined battery power
				'/Dc/Battery/Soc': dummy, # combined battery SoC
				'/Dc/Pv/Power': dummy, # combined PV power
				'/Relay/0/State': dummy, # MultiPlus cooling fans
			},
			'com.victronenergy.vebus': {
				'/Hub4/L1/AcPowerSetpoint': dummy, # setpoint L1
				'/Hub4/L2/AcPowerSetpoint': dummy, # setpoint L2
				'/Hub4/L3/AcPowerSetpoint': dummy, # setpoint L3
				'/Hub4/DisableCharge': dummy, # disable charging
				'/Hub4/DisableFeedIn': dummy, # disable inverters
				'/Hub4/DoNotFeedInOvervoltage': dummy, # feed excess to grid
			}
		}

		# start dbus monitor to get live updates when values change
		self._dbus_monitor = DbusMonitor(
			self._dbus_tree,
			valueChangedCallback=self._dbus_value_changed
		)
		self._ess_mode = self._dbus_monitor.get_value('com.victronenergy.settings', '/Settings/CGwacs/Hub4Mode')

	def _tick_grid(self, power):
		# do nothing if output is disabled
		if not self._outputEnabled:
			return
		
		# calculate optimal output value to keep grid setpoint
		output = self._get_setpoint() - power + GRID_SETPOINT

		# enforce output limits
		if output < self._outputLimits[0]:
			output = self._outputLimits[0]
		elif output > self._outputLimits[1]:
			output = self._outputLimits[1]
		
		# send result to inverters
		self._set_inverter_power(output)

	def _tick_battery(self):
		# set flag if battery reached minimum SoC
		soc = self._dbus_monitor.get_value('com.victronenergy.system', '/Dc/Battery/Soc')
		if not self._batteryLow and soc <= BATTERY_MIN_SOC:
			self._batteryLow = True
		elif self._batteryLow and soc >= BATTERY_MIN_SOC + BATTERY_THRESHOLD:
			self._batteryLow = False
		
		# switch modes and outputs
		now = time.time()
		pv = self._get_pv_power()
		setpoint = self._get_setpoint()
		if now - self._lastModeUpdate > MODE_UPDATE_DELAY:
			self._lastModeUpdate = now
			
			# control cooling fans
			if setpoint < -FAN_ON_POWER:
				self._lastFanStart = now
				self._dbus_monitor.set_value('com.victronenergy.system', '/Relay/0/State', 1)
			elif now - self._lastFanStart > FAN_OFF_DELAY:
				self._dbus_monitor.set_value('com.victronenergy.system', '/Relay/0/State', 0)

			# enable/disable feed to grid
			for service in self._dbus_monitor.get_service_list():
				if service.startswith('com.victronenergy.vebus') and service not in EXCLUDE_DEVICES:
					# allow over-voltage feed in if configured
					if self._dbus_monitor.get_value(service, '/Hub4/DoNotFeedInOvervoltage') != 0 if FEED_IN_EXCESS else 1:
						self._dbus_monitor.set_value(service, '/Hub4/DoNotFeedInOvervoltage', 0 if FEED_IN_EXCESS else 1)
					# always enable feed in
					if self._dbus_monitor.get_value(service, '/Hub4/DisableFeedIn') != 0:
						self._dbus_monitor.set_value(service, '/Hub4/DisableFeedIn', 0)
					# always enable charge
					if self._dbus_monitor.get_value(service, '/Hub4/DisableCharge') != 0:
						self._dbus_monitor.set_value(service, '/Hub4/DisableCharge', 0)
					break

			# always disable BatteryLife
			self._dbus_monitor.set_value('com.victronenergy.settings', '/Settings/CGwacs/BatteryLife/State', 10)

			if self._batteryLow and pv < PV_THRESHOLD:
				# low battery and no PV, disable inverters
				if self._outputEnabled:
					self._outputEnabled = False
					self._set_inverter_power(0)
			else:
				# battery or PV available, enable inverters
				self._outputEnabled = True

		# do not enfoce battery limits when inverters are off
		if not self._outputEnabled:
			return
		
		# calculate system efficiency
		sample = self._get_consumption() / (pv - self._get_battery_power())
		if sample > 0 and sample < 1:
			# put valid sample into averages list
			self._efficiency[self._efficiencyPos] = sample
			self._efficiencyPos += 1
			if self._efficiencyPos >= EFFICIENCY_SAMPLES:
				self._efficiencyPos = 0

		# handle low battery
		if self._batteryLow and pv > PV_THRESHOLD:
			# battery is low, slowly recharge it and use the rest for grid compensation
			limit = -((pv - BATTERY_LOW_CHARGE) * (sum(self._efficiency) / EFFICIENCY_SAMPLES))
			if limit > 0:
				limit = 0
			elif limit < -MAX_INVERTER_POWER:
				limit = -MAX_INVERTER_POWER
			self._outputLimits[0] = limit
		else:
			# battery normal, allow full power
			self._outputLimits[0] = -MAX_INVERTER_POWER

	def _dbus_value_changed(self, dbusServiceName, dbusPath, options, changes, deviceInstance):
		# update ESS mode if changed
		if dbusServiceName == 'com.victronenergy.settings' and dbusPath == '/Settings/CGwacs/Hub4Mode':
			self._ess_mode = changes['Value']

		# do nothing if ESS mode is not "external control"
		if self._ess_mode != 3:
			return

		# skip updates from excluded devices
		if dbusServiceName in EXCLUDE_DEVICES:
			return
		
		# check if this value change should trigger a control loop tick
		if dbusServiceName.startswith('com.victronenergy.grid') and dbusPath == '/Ac/Power':
			# grid meter value changed, execute grid calculations
			self._tick_grid(changes['Value'])
		elif dbusServiceName.startswith('com.victronenergy.system') and dbusPath == '/Dc/Battery/Power':
			# battery power changed, execute battery calculations
			self._tick_battery()

	def _get_pv_power(self):
		return self._dbus_monitor.get_value('com.victronenergy.system', '/Dc/Pv/Power')
	
	def _get_battery_power(self):
		return self._dbus_monitor.get_value('com.victronenergy.system', '/Dc/Battery/Power')
	
	def _get_consumption(self):
		return (
			self._dbus_monitor.get_value('com.victronenergy.system', '/Ac/Consumption/L1/Power') +
			self._dbus_monitor.get_value('com.victronenergy.system', '/Ac/Consumption/L2/Power') +
			self._dbus_monitor.get_value('com.victronenergy.system', '/Ac/Consumption/L3/Power')
		)
	
	def _get_setpoint(self):
		# returns the currently active grid setpoint of all inverters
		power = []
		for service in self._dbus_monitor.get_service_list():
			if service.startswith('com.victronenergy.vebus') and service not in EXCLUDE_DEVICES:
				for i in range(1, PHASE_COUNT + 1):
					power.append(self._dbus_monitor.get_value(service, '/Hub4/L%d/AcPowerSetpoint' % i))
		return sum(power)

	def _set_inverter_power(self, power):
		# sets the given setpoint (equal power on all inverters)
		for service in self._dbus_monitor.get_service_list():
			if service.startswith('com.victronenergy.vebus'):
				# set inverter power depending on phase count
				for i in range(1, PHASE_COUNT + 1):
					self._dbus_monitor.set_value(service, '/Hub4/L%d/AcPowerSetpoint' % i, int(power / PHASE_COUNT))

if __name__ == '__main__':
	# argument parsing
	parser = argparse.ArgumentParser(description='Custom ESS loop implementation in Python')
	parser.add_argument('-d', '--debug', help='set logging level to debug', action='store_true')
	args = parser.parse_args()
	logger = setup_logging(debug=args.debug)
	logger.info('-------- dbus-custom-ess, v' + SOFTWARE_VERSION + ' is starting up --------')

	# have a main loop, so we can send/receive asynchronous calls to and from dbus
	DBusGMainLoop(set_as_default=True)

	# start custom ESS
	dbus_custom_ess = DBusCustomESS()

	# run the main loop
	mainloop = GLib.MainLoop()
	mainloop.run()
